///////////////////////////////////////////////////////////////////////////////////////
//  EEPROM routines
#pragma region EEPROM

// address:
// 0  volume
// 1  input (0 based)
// 2  output (0 based)
// 3  33 marker
// 4  44 marker
// 5  VOL_MAX_STEP
// 6...  calibration data

#define OFFSET_L 6

void saveCalibration() {

  byte *p;

  PRINTLN("Save left");
  for (byte sample = 0; sample <= VOL_MAX_STEP - 3; sample++) {
    p = (byte*)&dataL[sample];
    for (byte i = 0; i < 6; i++) {  // LDRsample struct size = 6
      PRINT("addr "); PRINT(OFFSET_L + sample * 6 + i); PRINT(" byte "); PRINTLN(p[i]);
      if (EEPROM.read(OFFSET_L + sample * 6 + i) != p[i])
        EEPROM.write(OFFSET_L + sample * 6 + i, p[i]);
    }
  }

  PRINTLN("Save right");
  word offset_R = OFFSET_L + 6 * (VOL_MAX_STEP - 2);

  for (byte sample = 0; sample <= VOL_MAX_STEP - 3; sample++) {
    p = (byte*)&dataR[sample];
    for (byte i = 0; i < 6; i++) {  // LDRsample struct size = 6
      PRINT("addr "); PRINT(offset_R + sample * 6 + i); PRINT(" byte "); PRINTLN(p[i]);
      if (EEPROM.read(offset_R + sample * 6 + i) != p[i])
        EEPROM.write(offset_R + sample * 6 + i, p[i]);
    }
  }

  //write marker:
  EEPROM.write(3, 33); EEPROM.write(4, 44); EEPROM.write(5, VOL_MAX_STEP);
}


bool loadCalibration() {
  byte *p;
  word offset_R = OFFSET_L + 6 * (VOL_MAX_STEP - 2);

  if (EEPROM.read(3) != 33 || EEPROM.read(4) != 44 || EEPROM.read(5) != VOL_MAX_STEP)
    return false; //no data

  PRINTLN("Load left");
  for (byte sample = 0; sample <= VOL_MAX_STEP - 3; sample++) {
    p = (byte*)&dataL[sample];
    for (byte i = 0; i < 6; i++) {
      p[i] = EEPROM.read(OFFSET_L + sample * 6 + i);
      PRINT("addr "); PRINT(OFFSET_L + sample * 6 + i); PRINT(" byte "); PRINTLN(p[i]);
    }
  }

  PRINTLN("Load right");
  for (byte sample = 0; sample <= VOL_MAX_STEP - 3; sample++) {
    p = (byte*)&dataR[sample];
    for (byte i = 0; i < 6; i++) {
      p[i] = EEPROM.read(offset_R + sample * 6 + i);
      PRINT("addr "); PRINT(offset_R + sample * 6 + i); PRINT(" byte "); PRINTLN(p[i]);
    }
  }

  return true;
}


void saveIOValues() {
  if (EEPROM.read(0) != volume)
    EEPROM.write(0, volume);
  if (EEPROM.read(1) != chan_in)
    EEPROM.write(1, chan_in);
  if (EEPROM.read(2) != chan_out)
    EEPROM.write(2, chan_out);
}

void loadIOValues() {
  volume = EEPROM.read(0);
  chan_in = EEPROM.read(1);
  chan_out = EEPROM.read(2);
  if (volume > VOL_MAX_STEP)
    volume = VOL_DEFAULT;
  if (chan_in >= INPUTCOUNT)
    chan_in = 0;
  if (chan_out >= OUTPUTCOUNT)
    chan_out = 0;
}


//void eraseEEPROM() {
//  for (byte i = 0; i < 1024; i++) //1024 max
//    EEPROM.write(i, 255);
//}
#pragma endregion
