/* Rotary encoder handler for arduino.

   Copyright 2011 Ben Buxton. Licenced under the GNU GPL Version 3.
   Contact: bb@cactii.net

   Quick implementation of rotary encoder routine.

   More info: http://www.buxtronix.net/2011/10/rotary-encoders-done-properly.html
*/

// Arduino pins the encoder is attached to. Attach the center to ground.  - Already defined
//#define PIN_ENC1 2
//#define PIN_ENC2 3

// Pins should be setup like this (already done in our code setup
// define to enable weak pullups.
// #define ENABLE_PULLUPS
/* Call this once in setup(). */
/*void rotary_init() {
  pinMode(PIN_ENC1, INPUT);
  pinMode(PIN_ENC2, INPUT);
  #ifdef ENABLE_PULLUPS
  digitalWrite(PIN_ENC1, HIGH);
  digitalWrite(PIN_ENC2, HIGH);
  #endif
  }*/

// Half-step mode?
#define DIR_CCW 0x10
#define DIR_CW 0x20

#ifdef HALF_STEP
// Use the half-step state table (emits a code at 00 and 11)
const unsigned char ttable[6][4] = {
{0x3 , 0x2, 0x1,  0x0}, {0x23, 0x0, 0x1,  0x0},
{0x13, 0x2, 0x0,  0x0}, {0x3 , 0x5, 0x4,  0x0},
{0x3 , 0x3, 0x4, 0x10}, {0x3 , 0x5, 0x3, 0x20}
};
#else
// Use the full-step state table (emits a code at 00 only)
const unsigned char ttable[7][4] = {
{0x0, 0x2, 0x4, 0x0}, {0x3, 0x0, 0x1, 0x10},
{0x3, 0x2, 0x0, 0x0}, {0x3, 0x2, 0x1, 0x0},
{0x6, 0x0, 0x4, 0x0}, {0x6, 0x5, 0x0, 0x20},
{0x6, 0x5, 0x4, 0x0},
};
#endif


volatile unsigned char encoderState = 0;

/* Read input pins and process for events. Call this either from a
   loop or an interrupt (eg pin change or timer).

   Returns 0 on no event, otherwise 0x80 or 0x40 depending on the direction.
*/
unsigned char rotary_process() {
  unsigned char pinstate = (digitalRead(PIN_ENC2) << 1) | digitalRead(PIN_ENC1);
  encoderState = ttable[encoderState & 0xf][pinstate];
  return (encoderState & 0x30);
}

///////////////////////////////////////////////////////////////////////////////////////
//  Encoder routines
#pragma region Encoder
void attachInterrupts() {
  PCintPort::attachInterrupt(PIN_ENC2, &doEncoderB, CHANGE);
  attachInterrupt(0, doEncoderA, CHANGE);
}

void encoderHandler() {
  unsigned char result = rotary_process();
  if (result) {
    if (result == DIR_CCW)
      encoderPos++;
    else
      encoderPos--;
  }
}

void doEncoderA() {
  encoderHandler();
}

void doEncoderB() {
  encoderHandler();
}
#pragma endregion
