//#include <Wire.h> //
#include <PinChangeInt.h>
#include <EEPROM.h>
#include "Adafruit_MCP23008.h"
#include "LDRstruct.h"
// #include <SPI.h>  //
//#include "SSD1306Ascii.h"  //
#include "SSD1306AsciiWire.h"

SSD1306AsciiWire oled;

#pragma region defines

// Please only customize the values where the line comment starts with "//**" (unless you have a good reason)

// Debug mode increases memory usage, the code may not compile. In this case, reduce attenuation steps number to 30.
// !! COMMENT OUT THE FOLLOWING LINE WHEN NOT DEBUGGING ("//#define DEBUG") !!
// #define DEBUG

// Set I/O configuration here.
// There can be 0..2 inputs and 0..5 outputs, OR 0..5 inputs and 0..2 outputs.
// So for example we can have 2 IN and 5 OUT, or 3 IN and 2 OUT, or 4 IN and 1 OUT, but NOT 3 IN and 3 OUT.
#define INPUTCOUNT 5	 //** number of inputs, 0 to 5. If 0, comment out the next line.
const char* const inputName[INPUTCOUNT] = {"TURNTABLE", "SQUEZEBOX", "USB DAC 1", "USB DAC 2" , "MPLAYER"}; //** each name maximum 9 characters. There must be exactly INPUTCOUNT names in the list.

#define OUTPUTCOUNT 2   //** number of outputs, 0 to 4. If 0, comment out the next line.
const char* const outputName[OUTPUTCOUNT] = { "LINE OUT", "HEADPHONE" }; //** each name maximum 9 characters. There must be exactly OUTPUTCOUNT names in the list.

char msgWelcome1[] = "Gary Morgan's";                 //** <-- maximum 20 characters, depending on font (line 1)
char msgWelcome2[] = "VxD LDR Passive";               //** <-- maximum 20 characters, depending on font (line 2)
char msgWelcome3[] = "Optical PreAmp";                //** <-- maximum 20 characters, depending on font (line 3)
char msgWelcome4[] = "Version 2.0";                   //** <-- maximum 20 characters, depending on font (line 4)
char msgCalib[] = "Calibrating: ";                    //** <-- maximum 17 characters
char msgTest[] = "Self-testing";                      //** <-- maximum 17 characters
char msgBias[] = "Adjust BIAS";                       //** <-- maximum 19 characters
char msgMeas[] = "Measure LDRs";                      //** <-- maximum 19 characters
char msgCalibM[] = "Calibrate";                       //** <-- maximum 19 characters
char msgExit[] = "Exit";                              //** <-- maximum 19 characters
char msgErr[] = "ERROR ";                             //** <-- maximum 17 characters. Last char must be blank
char msgAbort[] = "Aborted.";                         //** <-- maximum 19 characters
char msgNoCalib[] = "Please calibrate";               //** <-- maximum 19 characters
#define menuFont X11fixed7x14
#define textFont Verdana12
#define runFont X11fixed7x14
#define bigFont modern
#define logoFont X11fixed7x14B      
#define I2C_ADDRESS 0x3C                              //Address of display

// If you need a delayed relay contact (for tube filament preheating, for example), uncomment the following #define line.
//#define DELAY 30  //** The value represents the delay in seconds. Maximum 100. Connects PIN_EXT_R6 to GND after DELAY seconds.


// Error codes:
// 1  : left series could not be calibrated
// 2  : right series could not be calibrated
// 3  : left shunt could not be calibrated
// 4  : right shunt could not be calibrated
// 10 : left series LDR too low max value. Adjust trimmer RT1
// 11 : right series LDR too low max value. Adjust trimmer RT3
// 12 : left shunt LDR too low max value. Adjust trimmer RT2
// 13 : right shunt LDR too low max value. Adjust trimmer RT4
// 20 : cannot set LDRs or calibration relays. Analog and/or relay power supplies are not connected?
// 30 : configuration error: INPUTCOUNT and OUTPUTCOUNT out of range


/******* Attenuator control *******/
#define NOM_IMPEDANCE 10000   //** target nominal attenuator impedance, between 5000 and 50000 Ohm. Recommended: 10000. Higher = less precision
#define LOAD_IMPEDANCE 220000 //** input impedance of the amplifier that follows the attenuator. Should not be less than 100K. Default: 220000 ohm
#define VOL_MAX_STEP 50       //** maximum volume steps; range: 20...80. Higher = more memory usage
#define VOL_DEFAULT 2         //** default volume step. Must be <= than MAX
#define volumeBar             //** turns on volume bar
#define AIO                   //** enables AIO (All-In-One) PCB, comment out for original PCB
#define BTN_HOLD_TIME 2000    //** Time needed to hold the menu button for entering setup menu. Default 2s
#define HALF_STEP             //** Uncomment for 1/2 step rotary encoders
#define TIME_EXITSELECT 3     //** Time in seconds to exit I/O select mode when no activity
#define RemoteIRTune 20000    //** Adjust this value to tune remote control operation (Default 25000)

/******* LDR measured values *******/
#define LDR_VOLTAGE 5.0       //** precisely measured value of the +5V supply (with decimal point. Default: 5.0)
#define LDR_R1 10000          //** precisely measured value of R1 resistor (default: 10000 ohm)
#define LDR_R18 10000         //** precisely measured value of R18 resistor (default: 10000 ohm)

// if any measured value is > 200 ohm, replace the LDR. Normal values are around 100 ohm.
#define LDR_LSE_MIN 61        //** measured value of left series LDR R at maximum current
#define LDR_LSH_MIN 111       //** measured value of left shunt LDR R at maximum current
#define LDR_RSE_MIN 81        //** measured value of right series LDR R at maximum current
#define LDR_RSH_MIN 111       //** measured value of right shunt LDR R at maximum current

//------- do not edit below this point ----------------------------------------------------//

#define TIME_IGNOREREMOTE_VOL  0    // Time in millisec after IR VOLUME before repeating (Default 100)
#define TIME_IGNOREREMOTE_CMD  0    // Time in millisec after IR command before repeating (Default 100)
#define TIME_SWITCHBOUNCE 0         // Time in millisec to debounce rotary switch (Default 250, 0 if using HW Implemented debounce)
#define TIME_IRBOUNCE 0             // Time in millisec to debounce IR button (Default 200)
#define TIME_DACSETTLE 50           // Time in millisec to wait for the DAC to stabilize after setting output (Default 500)
#define TIME_RELAYLATCH 0           // Time in millisec for latching I/O relay driver to stabilize (Default 250, 0 if not using latching relays)

/******* LDR control *******/
#define LDR_LOW_TOLERANCE 2 // percent between theory and measured R value acceptable for low R values
#define LDR_ZERO_MIN 300000 // minimum acceptable R when PWM=0 
#define LDR_FULL_MAX 200    // maximum acceptable R when PWM=255. 

/******* I/O PORTS *******/
#define PIN_ENC1 2        // Encoder 1
#define PIN_ENC2 4        // Encoder 2
#define PIN_BTN 5         // Encoder switch
#define PIN_LCDBRI 6      // LCD brightness control
#define PIN_REMOTE 12     // IR receiver (remote control) pin
#define PIN_DAC_LSE 11    // left channel series DAC pin
#define PIN_DAC_LSH 10    // left channel shunt DAC pin
#define PIN_DAC_RSE 9     // right channel series DAC pin
#define PIN_DAC_RSH 3     // right channel shunt DAC pin
#define PIN_SENSE_LSE A0  // left channel series sense pin
#define PIN_SENSE_LSH A6  // left channel shunt sense pin
#define PIN_SENSE_RSE A1  // right channel series sense pin
#define PIN_SENSE_RSH A7  // right channel shunt sense pin
#define PIN_BIAS_LSH 8    // left channel shunt bias pin
#define PIN_BIAS_RSH 7    // right channel shunt bias pin
#define PIN_LDR_L A2      // left channel LDR measure pin
#define PIN_LDR_R A3      // right channel LDR measure pin
#define PIN_EXT_CALIB 0   // port extender: calibration relays
#define PIN_EXT_R1 1      // port extender: relay 1
#define PIN_EXT_R2 2      // port extender: relay 2
#define PIN_EXT_R3 3      // port extender: relay 3
#define PIN_EXT_R4 6      // port extender: relay 4
#ifdef AIO                // swap R5 and R6 pins for ALL-IN-ONE controller PCB
#define PIN_EXT_R5 5      // port extender: relay 5
#define PIN_EXT_R6 4      // port extender: relay 6
#else if
#define PIN_EXT_R5 4      // port extender: relay 5
#define PIN_EXT_R6 5      // port extender: relay 6
#endif
#define PIN_EXT_BIAS 7    // port extender: BIAS

/****** IR remote codes ******/
#define cIR_UP 5
#define cIR_DOWN 6
#define cIR_LEFT 4
#define cIR_RIGHT 3
#define cIR_CENTER 46
#define cIR_MENU 1
#define cIR_PLAY 47

/******* INTERFACE ********/
// Selection order in normal mode
#define SEL_INPUT 0
#define SEL_OUTPUT 1

//Selection order in setup mode. You also need to edit code in "screen" tab
#define SEL_BIAS 1
#define SEL_MEAS 2
#define SEL_CALIB 3
#define SEL_EXIT 0

//#define B 255
//#define A 32

#if INPUTCOUNT == 0 && OUTPUTCOUNT == 0
#define VOLCOL 6
#else
#define VOLCOL 11
#endif

#ifdef DEBUG
#define PRINT(x) Serial.print(x)     // console logging
#define PRINTLN(x) Serial.println(x)
#else
#define PRINT(x)
#define PRINTLN(x)
#endif


/******* MACHINE STATES *******/
#define STATE_RUN 0        // normal run state
#define STATE_IO 1         // when user selects input/output
#define STATE_SETUP 2      // in setup menu mode
#define STATE_CALIB 3      // calibration and measuring 
#define STATE_ERROR 4
#define STATE_OFF 5
#define ON HIGH
#define OFF LOW

#pragma endregion defines


///////////////////////////////////////////////////////////////////////////////////////
// VARIABLES
#pragma region variables
/******* GLOBAL *******/
Adafruit_MCP23008 mcp;            // port extender
byte volume;                      // current volume, between 0 and VOL_STEPS
byte chan_in;                     // current input channel. 0 based
byte chan_out;                    // current output channel. 0 based
byte state;                       // current machine state
bool isMuted;			                // current mute status
byte selIO;	     		              // selected interface option in normal mode
byte selSetup;			              // selected interface option in setup mode
byte errc;			                  // error code
bool btnReleased;
bool btnPressed;
bool IRPressed;
byte LSHrange;
byte RSHrange;
unsigned int tick;
byte percent;
bool notCalibrated;		             // if no calibration data exists, remain in Setup mode
unsigned long calibStarted;


/******** LDR ********/
// LDR necessary current for each volume step.
// VOL_MAX_STEP + 1 total volume steps, including mute and full
// Step vol=0: MUTE, no data necessary
// Step vol=1: min volume (max series R + min shunt R), no data necessary
// Step vol=VOL_MAX_STEP: full volume (min series R + max shunt R), no data necessary
// So we have VOL_MAX_STEP+1 volume steps, but VOL_MAX_STEP-2 stored samples
// vol=2 -> data[0]
// vol=3 -> data[1]
LDRsample dataL[VOL_MAX_STEP - 2]; // left channel
LDRsample dataR[VOL_MAX_STEP - 2]; // right channel
int deltaSample;
char deltaPW;


/******* REMOTE *******/
unsigned long mil_onRemote = 0; //debounce remote

int duration;          // Duration of the IR pulse
int mask;              // Used to decode the remote signals
byte c1;               // Byte 1 of the 32-bit remote command code
byte c2;               // Byte 2 of the 32-bit remote command code
byte c3;               // Byte 3 of the 32-bit remote command code
byte c4;               // Byte 4 of the 32-bit remote command code
byte IRkey;            // The unique code (Byte 3) of the remote key
byte IRkey2;           // The unique code (Byte 4) of the remote key
byte previousIRkey;    // The previous code (used for repeat)
bool isIRrepeat;       // Repeating code
byte lastInput = 0;        // Last selected input for A/B switching button
byte prevInput = 0;
unsigned long mil_onRemoteKey = 0; // Stores time of last remote command

/******* ENCODER *******/
volatile int encoderPos = 0;       // encoder clicks


/******* TIMING *******/
unsigned long mil_onButton;    // Stores last time for switch debounce
unsigned long mil_onAction;    // Stores last time of user input
unsigned long mil_onSetLDR = 0;    // Stores last time that a LDR has changed value
unsigned long mil_onInput;     // Last time relay set
unsigned long mil_onOutput;    // Last time relay set
unsigned long mil_delta;
unsigned long mil_btnHold;     // Stores start time of button hold

#pragma endregion

///////////////////////////////////////////////////////////////////////////////////////
//  Pin modes
#pragma region pinModes
void setPinModes() {

  Wire.begin();
  mcp.begin();
  mcp.pinMode(PIN_EXT_CALIB, OUTPUT);
  setCalibrationRelays(ON);
  mcp.pinMode(PIN_EXT_R1, OUTPUT);
  mcp.pinMode(PIN_EXT_R2, OUTPUT);
  mcp.pinMode(PIN_EXT_R3, OUTPUT);
  mcp.pinMode(PIN_EXT_R4, OUTPUT);
  mcp.pinMode(PIN_EXT_R5, OUTPUT);
  mcp.pinMode(PIN_EXT_R6, OUTPUT);
  mcp.pinMode(PIN_EXT_BIAS, OUTPUT);

  pinMode(PIN_BIAS_LSH, INPUT);
  pinMode(PIN_BIAS_RSH, INPUT);
  digitalWrite(PIN_BIAS_LSH, LOW);      //Disable internal Pull Up
  digitalWrite(PIN_BIAS_RSH, LOW);
  LSHrange = RSHrange = HIGH;

  pinMode(PIN_SENSE_LSE, INPUT);
  pinMode(PIN_SENSE_LSH, INPUT);
  pinMode(PIN_SENSE_RSE, INPUT);
  pinMode(PIN_SENSE_RSH, INPUT);

  pinMode(PIN_LDR_L, INPUT);
  pinMode(PIN_LDR_R, INPUT);

  pinMode(PIN_ENC1, INPUT);
  pinMode(PIN_ENC2, INPUT);
  pinMode(PIN_BTN, INPUT_PULLUP);
  pinMode(PIN_REMOTE, INPUT_PULLUP);
  pinMode(PIN_DAC_LSE, OUTPUT);
  pinMode(PIN_DAC_LSH, OUTPUT);
  pinMode(PIN_DAC_RSE, OUTPUT);
  pinMode(PIN_DAC_RSH, OUTPUT);
  analogWrite(PIN_DAC_LSE, 0);
  analogWrite(PIN_DAC_LSH, 0);
  analogWrite(PIN_DAC_RSE, 0);
  analogWrite(PIN_DAC_RSH, 0);

  pinMode(PIN_LCDBRI, OUTPUT);
  analogWrite(PIN_LCDBRI, 0);
}
#pragma endregion

////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
//   SETUP
#pragma region setup
void setup() {
  Wire.begin();
  oled.begin(&Adafruit128x64, I2C_ADDRESS);
  oled.set400kHz();
  oled.clear();

  setPinModes();

  setLSE_Range(LOW); setRSE_Range(LOW); setLSE(255); setRSE(255);
  setLSH_Range(LOW); setRSH_Range(LOW); setLSH(255); setRSH(255);

  // set 31250Hz PWM frequency on pins 3,9,10,11
  TCCR1B = TCCR1B & 0b11111000 | 0x01;
  TCCR2B = TCCR2B & 0b11111000 | 0x01;


#ifdef DEBUG
Serial.begin(57600);
#endif

  // display welcome message
  oled.setFont(logoFont);
  oled.setCursor((128 - sizeof(msgWelcome1) * oled.fontWidth()) / 2, 0);
  oled.println (msgWelcome1);
  oled.setCursor((128 - sizeof(msgWelcome2) * oled.fontWidth()) / 2, 2);
  oled.println (msgWelcome2);
  oled.setCursor((128 - sizeof(msgWelcome3) * oled.fontWidth()) / 2, 4);
  oled.println (msgWelcome3);
  oled.setCursor((128 - sizeof(msgWelcome4) * oled.fontWidth()) / 2, 6);
  oled.print (msgWelcome4);
  delay(2000);

  //test if the relays and LDRs are powered
  if (getRLSE() + getRRSE() + getRLSH() + getRRSH() > LDR_FULL_MAX * 4) {
    errc = 20;
    toErrorState();
  }
  else if ((INPUTCOUNT > 5 && OUTPUTCOUNT > 2) || (INPUTCOUNT > 2 && OUTPUTCOUNT > 4) || (INPUTCOUNT == 3 && OUTPUTCOUNT == 3)) {
    errc = 30;
    toErrorState();
  }
  else
  {
    notCalibrated = !loadCalibration();
    if (notCalibrated)
      toSetupState();

    else {
#ifdef DELAY
setLSE(50); setRSE(50);
setLSH(50); setRSH(50);
doDelay();
#endif
loadIOValues();
// state = STATE_IO;
oled.clear();
drawRunDisplay();
setInput();
setOutput();
isMuted = volume == 0;
setVolume(volume);
delay(250);
setMute(volume);
toRunState();
}
}


      //         mil_onAction = millis(); // begin counting to dim the LCD

      mcp.digitalWrite(PIN_EXT_BIAS, HIGH); // turn biasing voltage on

      btnReleased = true;
      btnPressed = false;
      IRPressed = false;
      attachInterrupts(); // start reading encoder
      encoderPos = 0;
    }
    #pragma endregion


    //////////////////////////////////////////////////////////////////////////////////////
    // RELAYS and I/O control
    #pragma region Relays

    byte relayMap[6] = {PIN_EXT_R1, PIN_EXT_R2, PIN_EXT_R3, PIN_EXT_R4, PIN_EXT_R5, PIN_EXT_R6};

    void setCalibrationRelays(byte val) {
      mcp.digitalWrite(PIN_EXT_CALIB, val);
      isMuted = val > 0;
    }

    void setMute(byte volume) {
      isMuted = volume == 0;
      mcp.digitalWrite(PIN_EXT_CALIB, isMuted);
      drawVolume(volume);
    }

    void setInput() {
      bool wasMuted = isMuted;

#if INPUTCOUNT > 0
if (millis() - mil_onInput > TIME_RELAYLATCH) {
//    setMute(0);

      drawInput();

#if INPUTCOUNT > 2 // inputs are on R1 to R5
for (byte i = 0; i < INPUTCOUNT; i++)
mcp.digitalWrite(relayMap[i], (chan_in == i));

#elif INPUTCOUNT <= 2 && OUTPUTCOUNT <= 2 // inputs are on R1
mcp.digitalWrite(PIN_EXT_R1, chan_in);

#else // inputs are on R6
mcp.digitalWrite(PIN_EXT_R6, chan_in);

#endif
mil_onInput = millis();
}

#endif

      if (isMuted && !wasMuted)
        setVolume(volume);
    }

    void setOutput() {

#if	OUTPUTCOUNT > 0
//  strOffset = 102 - (strlen(outputName[chan_out]) + 1) * 6;
if (millis() - mil_onOutput > TIME_RELAYLATCH) {

      drawOutput();

#if OUTPUTCOUNT > 2 // outputs are on R1 to R5
for (byte i = 0; i < OUTPUTCOUNT; i++)
mcp.digitalWrite(relayMap[i], (chan_in == i));
#else // outputs are on R6
mcp.digitalWrite(PIN_EXT_R6, (chan_out));
#endif
mil_onOutput = millis();
}
#endif
}

      #pragma endregion


      ///////////////////////////////////////////////////////////////////////////////////////
      //  Delayed startup
      #pragma region delay
#ifdef DELAY
void doDelay() {

      // Turns on R6 relay after a DELAY number of seconds
      for (int i = DELAY; i > 0; i--)	{
        drawVolume(i);
        delay(1000);
      }
      mcp.digitalWrite(PIN_EXT_R5, HIGH);

    }
#endif
#pragma endregion


    //////////////////////////////////////////////////////////////////////////////////////
    // IR Remote
    #pragma region Remote
    /*
      IR code by Hifiduino. We cannot use _cmd lib because of timers interference.

      The following function returns the code from the Apple Aluminum remote control. The Apple remote is
      based on the NEC infrared remote protocol. Of the 32 bits (4 bytes) coded in the protocol, only the
      third byte corresponds to the keys. The function also handles errors due to noise (returns 255) and
      the repeat code (returs zero)

      The Apple remote returns the following codes:

      Up key:     238 135 011 089
      Down key:   238 135 013 089
      Left key:   238 135 008 089
      Right key:  238 135 007 089
      Center key: 238 135 093 089 followed by 238 135 004 089 (See blog for why there are two codes)
      Menu key:   238 135 002 089
      Play key:   238 135 094 089 followed by 238 135 004 089

      (update) The value of the third byte varies from remote to remote. It turned out that the 8th bit
      is not part of the command, so if we only read the 7 most significant bits, the value is the same
      for all the remotes, including the white platic remote.

      The value for the third byte when we discard the least significant bit is:

      Up key:     238 135 005 089
      Down key:   238 135 006 089
      Left key:   238 135 004 089
      Right key:  238 135 003 089
      Center key: 238 135 046 089 followed by 238 135 002 089 (See blog for why there are two codes)
      Menu key:   238 135 001 089
      Play key:   238 135 047 089 followed by 238 135 002 089

      More info here: http://hifiduino.wordpress.com/apple-aluminum-remote/
    */

    int getIRkey() {

      c1 = c2 = c3 = c4 = 0;
      duration = 1;
      bool repeat = false;

      while ((duration = pulseIn(PIN_REMOTE, HIGH, RemoteIRTune)) < 2000 && duration != 0)
      {
        // Wait for start pulse
      }
      PRINTLN(duration);

      if (duration == 0) 							// This is an error no start or end of pulse
        return (255);							    // Use 255 as Error

      if (duration > 1000) {					// This is the repeat
        PRINT("repeat\n");
        repeat = true;
      }

      if (duration < 5000) {					// This is the command get the 4 byte
        mask = 1;
        for (int i = 0; i < 8; i++) {					// get 8 bits
          if (pulseIn(PIN_REMOTE, HIGH, 3000) > 1000)  // If "1" pulse
            c1 |= mask;					// Put the "1" in position
          mask <<= 1;						// shift mask to next bit
        }
        mask = 1;
        for (int i = 0; i < 8; i++) {					// get 8 bits
          if (pulseIn(PIN_REMOTE, HIGH, 3000) > 1000)  // If "1" pulse
            c2 |= mask;					// Put the "1" in position
          mask <<= 1;						// shift mask to next bit
        }
        mask = 1;
        for (int i = 0; i < 8; i++) {					// get 8 bits
          if (pulseIn(PIN_REMOTE, HIGH, 3000) > 1000)  // If "1" pulse
            c3 |= mask;					// Put the "1" in position
          mask <<= 1;      					// shift mask to next bit
        }
        mask = 1;
        for (int i = 0; i < 8; i++) {         // get 8 bits
          if (pulseIn(PIN_REMOTE, HIGH, 3000) > 1000)  // If "1" pulse
            c4 |= mask;         // Put the "1" in position
          mask <<= 1;               // shift mask to next bit
        }

        c3 >>= 1;   // Discard the least significant bit

        PRINT("c1:"); PRINT(c1); PRINT(" c2:"); PRINT(c2); PRINT(" c3:"); PRINT(c3); PRINT(" c4:"); PRINT(c4); PRINTLN();
        if ((c1 == 238) && (c2 == 135) /*&& c4 == 131*/) {                // Only return valid Apple remote codes
          PRINTLN(c3);
          return (c3);
        }

        if (repeat && (c3 == 2))                                          // Only way to verify a repeat from this remote is with the 2 code?
          PRINT("repeat\n");                                              // Code seems to need soemthing on this line to repeat??
          return (0);
      }
      return (255);
    }
    #pragma endregion


    ///////////////////////////////////////////////////////////////////////////////////////
    //  state transitions
    #pragma region Transitions

    /** transition to setup mode, display menu **/


    void toSetupState() {

      //selSetup = SEL_BIAS;
      selSetup = SEL_EXIT;
      oled.clear();
      setupMenu();

      state = STATE_SETUP;
    }

    /** transition to error state **/
    void toErrorState() {
      setCalibrationRelays(ON);
      setLSE(0); setRSE(0);
      setLSH(0); setRSH(0);
      oled.clear();
      oled.setCursor(36, 2);
      oled.setFont(textFont);
      oled.print(msgErr);
      oled.print(errc);
      state = STATE_ERROR;
    }

    /** transition to I/O channel adjust mode **/
    void toIOState() {
      state = STATE_IO;
      //drawRunDisplay();
    }
    // remember last two channel inputs for A/B switching

    void storeLast() {
      if (chan_in != lastInput) {
        prevInput = lastInput;
        lastInput = chan_in;
        PRINT("chan:"); PRINT(chan_in); PRINT(" prev:"); PRINT(prevInput); PRINT(" LAST:"); PRINT(lastInput); PRINTLN();
      }
    }

    /** transition to normal volume adjust mode **/
    void toRunState() {
      oled.setCursor(26, 0);
      oled.print(">");
      oled.setCursor(98, 2);
      oled.print(">");
      selIO = SEL_INPUT;
      state = STATE_RUN;
      //drawRunDisplay();
      //setVolume(volume);
      mil_onAction = millis();
    }

    #pragma endregion

    ////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////
    //         LOOP
    ////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////
    #pragma region loop
    void loop() {

      #pragma region powerOff
      // Detect power off
      if (state == STATE_RUN) {
        long vcc;
        ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
        delay(2);
        ADCSRA |= _BV(ADSC);
        while (bit_is_set(ADCSRA, ADSC));
        vcc = ADCL;
        vcc |= ADCH << 8;
        vcc = 1126400L / vcc;
        if (vcc > 3000 && vcc < 4600) {
          PRINT("power down...\n");                                // ***
          mcp.digitalWrite(PIN_EXT_BIAS, LOW);
          setLSE(0); setRSE(0);
          setLSH_Range(LOW); setRSH_Range(LOW);
          setLSH(255); setRSH(255);
          saveIOValues();
          state = STATE_OFF;
        }
      }
      #pragma endregion


      /////////////////////////////////////////////////////////////////////////////////////
      // IR Remote
      #pragma region loop_remote
      if (millis() - mil_onRemote > TIME_IGNOREREMOTE_CMD && digitalRead(PIN_REMOTE) == LOW && (state == STATE_RUN || state == STATE_IO)) {

        IRkey = getIRkey();
        if (IRkey != 255 && IRkey != 2) {

          mil_onAction = millis();
          isIRrepeat = IRkey == 0;

          // Prevent repeating if a code has not been received for a while.
          if ((millis() - mil_onRemoteKey) > 500)  {
            isIRrepeat = 0;
          }

          if (isIRrepeat && (previousIRkey == cIR_UP || previousIRkey == cIR_DOWN))  // Repeat the specified keys
            IRkey = previousIRkey;
          else
            previousIRkey = IRkey;

          mil_onRemoteKey = millis();

          PRINT("IR: "); PRINTLN(IRkey);

          switch (IRkey) {
            case cIR_UP:  // volume up
              encoderPos++;
              break;
            case cIR_DOWN:  // volume down
              encoderPos--;
              break;
            case cIR_LEFT:
              if (INPUTCOUNT > 1) {
                chan_in = chan_in - 1 + (chan_in == 0) * INPUTCOUNT;
                storeLast();
                setInput();
              }
              break;
            case cIR_RIGHT:
              if (INPUTCOUNT > 1) {
                //toIOState();
                //encoderPos++;
                chan_in++;
                chan_in %= INPUTCOUNT;
                storeLast();
                setInput();
              }
              break;
            case cIR_PLAY:
              toggleMute();
              break;
            //    case cIR_CENTER:
            ////      if (lastInput == prevInput)break;
            ////      chan_in = prevInput;
            ////      storeLast();
            ////      setInput();
            //      break;
            case cIR_MENU:
              if (state == STATE_IO) {
                toRunState();
                break;
              }
              else {
                if (OUTPUTCOUNT > 1) {
                  chan_out++;
                  chan_out = chan_out % OUTPUTCOUNT;
                  setOutput();
                }
                break;
              }
          } // end SWITCH

          if (!isIRrepeat) {
            mil_onRemote = millis();
          }
        }

      }
      else
        IRkey = 255;
      #pragma endregion



      /////////////////////////////////////////////////////////////////////////////////////
      // Rotary encoder was turned
      #pragma region loop_encoder
      //encoderRotating = true;
      if (encoderPos != 0)
      {
        mil_onAction = millis();
        //		startLCDFadeIn();


        /** encoder rotated in volume mode **/
        if (state == STATE_RUN) {
          volume += encoderPos;
          if (volume > 250)	// byte overflow...
            volume = 0;
          if (volume > VOL_MAX_STEP)
            volume = VOL_MAX_STEP;
          setVolume(volume);
        }


        /** encoder rotated in channel select mode **/
        else if (state == STATE_IO) {
          switch (selIO % 2) {
            case SEL_INPUT:
              if (INPUTCOUNT > 0) {
                chan_in += encoderPos;
                /*if (encoderPos > 0 )
                  chan_in++;
                  if (encoderPos < 0)
                  chan_in--;*/
                if (chan_in == 255) {
                  chan_in = INPUTCOUNT - 1;
                  //              storeLast();
                }
                else {
                  chan_in %= INPUTCOUNT;
                  //              storeLast();
                }
                setInput();
              }
              break;

            case SEL_OUTPUT:
              if (OUTPUTCOUNT > 0) {
                chan_out += encoderPos;
                if (chan_out == 255)
                  chan_out = OUTPUTCOUNT - 1;
                else
                  chan_out %= OUTPUTCOUNT;
                setOutput();
              }
          }
        }


        /** encoder rotated in "setup" mode **/
        else if (state == STATE_SETUP) {
          selSetup += encoderPos;
          selSetup %= 4;
          setupMenu();
        }
        encoderPos = 0; // Reset the flag
      }  // End of encoder rotation
      #pragma endregion


      //////////////////////////////////////////////////////////////////////////////////////
      // Button press
      //////////////////////////////////////////////////////////////////////////////////////
      #pragma region loop_button
      btnPressed = digitalRead(PIN_BTN) == LOW && btnReleased;
      //IRPressed = IRkey == cIR_CENTER && millis() - mil_onButton > TIME_IRBOUNCE;

      ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      if (((millis() - mil_btnHold) > BTN_HOLD_TIME) && !btnReleased && state == STATE_RUN) {     // hold mod ///////////
        toSetupState();                                                                           // hold mod ///////////
      }                                                                                           // hold mod ///////////
      ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
      ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

      if (btnPressed || IRPressed) {
        mil_btnHold = millis ();                                      // hold mod
        btnReleased = false;
        mil_onButton = mil_onAction = millis();                       // Start debounce timer

        /** button pressed in normal mode **/
        if (state == STATE_RUN) {
          //          if (INPUTCOUNT <= 1 && OUTPUTCOUNT <= 1) {
          //            if (btnPressed) {
          //              toSetupState();
          //            }
          //          }
          //          else {
          oled.setFont(runFont);
          oled.setCursor(26, 0);
          oled.print(char(127));
          toIOState();
          setMute(0);
        }
        /** button pressed in IO mode **/
        else if (state == STATE_IO) {
          if (selIO == SEL_INPUT && OUTPUTCOUNT > 1) {
            selIO = SEL_OUTPUT;
            oled.setFont(runFont);
            oled.setCursor(26, 0);
            oled.print(">");
            oled.setCursor(98, 2);
            oled.print(char(127));
            toIOState();
          }
          else {
            storeLast();
            oled.setFont(runFont);
            oled.setCursor(26, 0);
            oled.print(">");
            oled.setCursor(98, 2);
            oled.print(">");
            toRunState();
          }

        }

        /** button pressed in setup menu **/
        else if (btnPressed && state == STATE_SETUP) {
          switch (selSetup) {
            case SEL_BIAS:
              setMaxResistanceMode();
              break;
            case SEL_MEAS:
              setMinResistanceMode();
              break;
            case SEL_CALIB:
              errc = doCalibration();
              if (errc == 100) {
                oled.setCursor (20, 20);
                oled.clear();
                oled.setFont(textFont);
                oled.print(msgAbort);
                delay(3000);
                toRunState();
                setVolume(volume);
                setCalibrationRelays(OFF);
              }
              else if (errc > 0) {
                toErrorState();
              }
              break;
            case SEL_EXIT:
              selSetup = SEL_BIAS;
              if (!notCalibrated) {
                oled.clear();
                toRunState();
                drawRunDisplay();
              }
              else {
                oled.setCursor (20, 3);
                oled.clear();
                oled.setFont(textFont);
                oled.print(msgNoCalib);
                delay(1000);
                toSetupState();
              }
              break;
          }
        }

        /** button pressed in adjust mode **/
        else if (btnPressed && state == STATE_CALIB)
          toSetupState();
      }
      #pragma endregion



      /////////////////////////////////////////////////////////////////////////////////////
      // timing
      #pragma region loop_timing
      //** timeout selection/menu
      if ((state == STATE_IO) && millis() - mil_onAction > TIME_EXITSELECT * 1000) {
        saveIOValues();  //btnpress after modifying I/O, save to EEPROM
        oled.setFont(textFont);
        toRunState();
      }


      //** encoder button debounce/release
      if (digitalRead(PIN_BTN) == HIGH && millis() - mil_onButton > TIME_SWITCHBOUNCE) {
        btnReleased = true;
        //mil_onButton = millis();
      }
      #pragma endregion


      /////////////////////////////////////////////////////////////////////////////////////
      // feedback to maintain LDR current
      #pragma region loop_feedback
      if ((millis() - mil_onSetLDR > TIME_DACSETTLE) && state == STATE_RUN && volume < VOL_MAX_STEP && volume > 1) {

        // LSE
        word sample = getILSE();
        deltaSample = int(dataL[volume - 2].i_SE & 32767) - sample;
        deltaPW = deltaSample > 0 ? 1 : -1;
        if (deltaPW * deltaSample > abs(deltaPW * getDeltaI(dataL[volume - 2].i_SE) - deltaSample) && dataL[volume - 2].pw_SE > 0 && dataL[volume - 2].pw_SE < 255) {
          dataL[volume - 2].pw_SE += deltaPW;
          setLSE(dataL[volume - 2].pw_SE);
        }


        // RSE
        sample = getIRSE();
        deltaSample = int(dataR[volume - 2].i_SE & 32767) - sample;
        deltaPW = deltaSample > 0 ? 1 : -1;
        if (deltaPW * deltaSample > abs(deltaPW * getDeltaI(dataR[volume - 2].i_SE) - deltaSample) && dataR[volume - 2].pw_SE > 0 && dataR[volume - 2].pw_SE < 255) {
          dataR[volume - 2].pw_SE += deltaPW;
          setRSE(dataR[volume - 2].pw_SE);
        }


        // LSH
        sample = getILSH();
        deltaSample = int(dataL[volume - 2].i_SH & 32767) - sample;
        deltaPW = deltaSample > 0 ? 1 : -1;
        if (deltaPW * deltaSample > abs(deltaPW * getDeltaI(dataL[volume - 2].i_SH) - deltaSample) && dataL[volume - 2].pw_SH > 0 && dataL[volume - 2].pw_SH < 255) {
          dataL[volume - 2].pw_SH += deltaPW;
          setLSH(dataL[volume - 2].pw_SH);
        }


        // RSH
        sample = getIRSH();
        deltaSample = int(dataR[volume - 2].i_SH & 32767) - sample;
        deltaPW = deltaSample > 0 ? 1 : -1;
        if (deltaPW * deltaSample > abs(deltaPW * getDeltaI(dataR[volume - 2].i_SH) - deltaSample) && dataR[volume - 2].pw_SH > 0 && dataR[volume - 2].pw_SH < 255) {
          dataR[volume - 2].pw_SH += deltaPW;
          setRSH(dataR[volume - 2].pw_SH);
        }

        mil_onSetLDR = millis();
      }
      #pragma endregion

    }
    #pragma endregion



